module casc
 implicit none
 character*256 output_file, decay_file, init_file, rel_path
 character*20  element_names(0:200)
 double precision time_cut, lambda_cut, it_probability, kt, effective_decay_rate(5), max_kt
 integer number_reaction_types, number_levels, number_isomers, number_intermediate
 integer a, z, n																	! A, Z, N of isotope under investigation
 integer,          dimension (:)     ,allocatable,target :: z_product			    ! Charge of reaction product nucleus
 integer,          dimension (:)     ,allocatable,target :: n_product			    ! Neutron number of reaction product nucleus
 double precision, dimension (:)     ,allocatable,target :: e_product			    ! Energy level in product nucleus
 double precision, dimension (:)     ,allocatable,target :: e_level			    	! Energy level of state (keV)
 double precision, dimension (:)     ,allocatable,target :: g_level			    	! degeneracy of level
 double precision, dimension (:)     ,allocatable,target :: j_level			    	! angular momentum J of level (h_bar)
 double precision, dimension (:)     ,allocatable,target :: p_level			    	! parity of level (+-1)
 double precision, dimension (:)     ,allocatable,target :: lambda			    	! terrestrial decay constant of state (1/s) 
 double precision, dimension (:,:)   ,allocatable,target :: transition				! terrestrial probability of internal transition as a function of initial level and final level
 double precision, dimension (:,:)   ,allocatable,target :: decay					! terrestrial probability of decay as a function of initial level and reaction channel
 double precision, dimension (:,:)   ,allocatable,target :: lambda_transition		! terrestrial internal transition rate as a function of initial level and final level
 double precision, dimension (:,:)   ,allocatable,target :: lambda_transition_wk	! Weisskopf prediction of terrestrial internal transition rate as a function of initial level and final level
 double precision, dimension (:,:)   ,allocatable,target :: lambda_decay			! terrestrial decay rate as a function of initial level and reaction channel
 double precision, dimension (:,:)   ,allocatable,target :: lambda_kt				! thermal excitation rate as a function of initial level and final level
 double precision, dimension (:,:)   ,allocatable,target :: lambda_transition_effective		! effective stellar internal transition rate as a function of initial and final long-lived (!!) level
 double precision, dimension (:,:)   ,allocatable,target :: lambda_decay_effective			! effective stellar decay rate as a function of initial long-lived level and reaction channel
 integer,          dimension (:,:)   ,allocatable,target :: output_unit_decay	    ! unit to which the output is written to
 integer,          dimension (:,:)   ,allocatable,target :: output_unit_transition  ! unit to which the output is written to
 logical wk_completion
 integer unit_counter
!
!
end module casc

program make_casc
 use casc
 implicit none
 integer kt_int, i
!
 wk_completion = .false.
 call initfiles 
 call initdecay 
 do i=1,2
   if (wk_completion) call init_wk
   call initoutput
   do kt_int=1,1000
     kt = kt_int
     call calculate_rates
     call output
   end do  
   do kt_int=110,100,10
     kt = kt_int
     call calculate_rates
     call output
   end do  
   print*,'Maximum temperature with valid assumptions: ', max_kt, ' keV'
   call close_files
   wk_completion = .not. wk_completion
 end do  
end 

subroutine calculate_effective_decay_rate
 use casc
 implicit none
 integer i,j, s, l
 double precision lls, lsl, lbs, lbl, a1, p, q, d,x1,x2
!
 effective_decay_rate = 0.
 if (number_isomers .ge.2) then
   s = 0                                    ! short
   l = 0									! long
   do i=1,number_levels
     if (lambda(i) .lt. lambda_cut) then   			! isomer/gs
       if (s.gt.0) then
         if (lambda(i) .lt. lambda(s)) then
           l=i
         else
           l=s
           s=i
         end if
         exit
       else
         s=i
       end if
     end if          
   end do
 end if
 lsl = lambda_transition_effective(s,l)
 lls = lambda_transition_effective(l,s)
 lbl = sum(lambda_decay_effective(l,:))
 lbs = sum(lambda_decay_effective(s,:))
 a1  = lls/lsl
! print*,'s,l,lsl,lls,lbl,lbs',s,l,lsl,lls,lbl,lbs
! 
! rough approximation
!
 effective_decay_rate(1) = lbl+lls*lbs/(lbs+lsl)
 effective_decay_rate(2) = (lbs*a1+lbl)/(1+a1)
 effective_decay_rate(3) = min(effective_decay_rate(1),effective_decay_rate(2))
!
! exact solution of differential equation
! 
 p = lsl+lbs+lls+lbl
 q = (lsl+lbs)*(lls+lbl)-lls*lsl
!
! eigenvalues are the solution of the equation x*x+p*x+q=0
! 
 d = p*p/4.d0 -q
 if (d.lt.0) then
   print*,'no solution to DGL', kT
 else
   d  = dsqrt(d)
   x1 = -p/2.d0 + d    
   x2 = -p/2.d0 - d    
 end if
!
!  we want -lambda instead of Eigenvalue
! 
 effective_decay_rate(4) = min(-x1,-x2)
 effective_decay_rate(5) = max(-x1,-x2)
 
end


subroutine output
 use casc
 implicit none
 integer i,j, uc
 logical below_cut
!
 below_cut = .true.
 do i=1, number_levels
   do j=1, number_levels
     if (lambda(i) .ge. lambda_cut) then                              ! short-lived state
       below_cut = below_cut .and. ( 10.*lambda_kt(i,j) .le. lambda(i))
!       if (10.*lambda_kt(i,j) .gt. lambda(i)) print*,kt,i,j
     end if  
   end do
 end do
 if (below_cut) then
   if (kt .gt. max_kt) max_kt = kt
 end if
 do i=1, number_levels
   if (lambda(i) .lt. lambda_cut) then   			! isomer/gs
     do j=1, number_reaction_types         
       uc = output_unit_decay(i,j)
!       if (below_cut) then
         write(uc, *) kT, lambda_decay_effective(i,j)
!       else
!         write(uc, *) kT, lambda_decay_effective(i,j), &
!         'Warning: Model might not be applicable'
!       end if  
     end do
     do j=1, number_levels         
       if (lambda(j) .lt. lambda_cut .and. i.ne.j) then   			! isomer/gs
         uc = output_unit_transition(i,j)
!         if (below_cut) then
           write(uc, *) kT, lambda_transition_effective(i,j)
!         else
!           write(uc, *) kT, lambda_transition_effective(i,j), &
!           'Warning: Model might not be applicable'
!         end if  
       end if  
     end do
   end if  
 end do
 uc = output_unit_decay(0,0)
 write(uc, *) kT, effective_decay_rate(3), dlog(2.d0)/effective_decay_rate(3), effective_decay_rate(4), effective_decay_rate(5)
end     

subroutine close_files
 use casc
 implicit none
 integer i
!
 do i=20,unit_counter
   close(i)
 end do
end 

subroutine initoutput
 use casc
 implicit none
 integer i,j
 character*3 c3
 character*7 c_end
!
 unit_counter=20
 c_end='.dat'
 if (wk_completion) c_end = '_wk' //trim(c_end)
!
! decays
!
 do i=1, number_levels
   if (lambda(i) .lt. lambda_cut) then   			! isomer/gs
     do j=1, number_reaction_types         
        write(c3,'(I3.3)') a				
        output_file = trim(rel_path) // c3 //trim(element_names(z))
        write(c3,'(I3.3)') i				
        output_file = trim(output_file) // '_level_'//c3
        write(c3,'(I3.3)') z_product(j)+n_product(j)				
        output_file = trim(output_file) // '_to_'// c3 // trim(element_names(z_product(j))) &
                      // '_effective_decay' // trim(c_end)
        print*,'Output file: ',trim(output_file)
        open(unit=unit_counter, file=output_file, status='unknown')
        write(unit_counter, *) 'kT (keV)             effective decay rate (1/s)'
        output_unit_decay(i,j) = unit_counter
        unit_counter=unit_counter+1
     end do
   end if  
 end do
!
! internal transitions
!
 do i=1, number_levels
   if (lambda(i) .lt. lambda_cut) then   			! isomer/gs
     do j=1, number_levels         
       if (lambda(j) .lt. lambda_cut .and. i.ne.j) then   			! isomer/gs
         write(c3,'(I3.3)') a				
         output_file = trim(rel_path) // c3 //trim(element_names(z))
         write(c3,'(I3.3)') i				
         output_file = trim(output_file) // '_level_'//c3
         write(c3,'(I3.3)') j				
         output_file = trim(output_file) // '_to_level_'// c3 // '_effective_transition' // trim(c_end)
         print*,'Output file: ',trim(output_file)
         open(unit=unit_counter, file=output_file, status='unknown')
         write(unit_counter, *) 'kT (keV)             effective transition rate (1/s)'
         output_unit_transition(i,j) = unit_counter
         unit_counter=unit_counter+1
       end if  
     end do
   end if  
 end do
!
!  effective lifetime of isotope
! 
 write(c3,'(I3.3)') a				
 output_file = trim(rel_path) // c3 //trim(element_names(z)) // '_effective_lifetime' // trim(c_end)
 print*,'Output file: ',trim(output_file)
 open(unit=unit_counter, file=output_file, status='unknown')
 write(unit_counter, *) 'kT (keV)             effective decay rate of isotope (1/s)    effective half life of isotope (s)'
 output_unit_decay(0,0) = unit_counter
 
end

subroutine calculate_rates
 use casc
 implicit none
 integer i,j,k,l
 double precision ll, sum_k, sum_l
!
 call calculate_thermal_excitation 
!
! effective transitions between long-lived states
!
 lambda_transition_effective = 0.
 do i=1, number_levels
   if (lambda(i) .lt. lambda_cut) then   			! isomer/gs
     do j=1, number_levels          				
       if (lambda(j) .lt. lambda_cut .and. j.ne.i) then  		! isomer/gs
         ll = lambda_kT(i,j) + lambda_transition(i,j)
         sum_l=0.
         do l=1, number_levels
           if (lambda(l) .ge. lambda_cut) then   			! intermediate state
             sum_k=0.
             do k=1,l-1
               if (lambda(k) .ge. lambda_cut) then   			! intermediate state
                 sum_k = sum_k + transition(l,k)*transition(k,j)
               end if
             end do
             sum_l = sum_l + lambda_kt(i,l)*(transition(l,j)+sum_k) 
           end if 
         end do
         lambda_transition_effective(i,j) = ll+sum_l
       end if
     end do
   end if
 end do  
 
!
! effective decays of long-lived states
!
 lambda_decay_effective = 0.
 do i=1, number_levels
   if (lambda(i) .lt. lambda_cut) then   			! isomer/gs
     do j=1, number_reaction_types          				
       ll = lambda_decay(i,j)
       sum_l=0.
       do l=1, number_levels
         if (lambda(l) .ge. lambda_cut) then   			! intermediate state
           sum_k=0.
           do k=1,l-1
             if (lambda(k) .ge. lambda_cut) then   			! intermediate state
               sum_k = sum_k + transition(l,k)*decay(k,j)
             end if
           end do
           sum_l = sum_l + lambda_kt(i,l)*(decay(l,j)+sum_k) 
         end if 
       end do
       lambda_decay_effective(i,j) = ll+sum_l
     end do
   end if
 end do  
 call calculate_effective_decay_rate 
end

subroutine calculate_thermal_excitation
 use casc
 implicit none
 integer i,j
!
 lambda_kt = 0.
 do i=1,number_levels
   do j=1,number_levels
     if (lambda_transition(i,j) .gt. 0.) then
       lambda_kt(j,i) = lambda_transition(i,j) * g_level(i)/g_level(j)
       lambda_kt(j,i) = lambda_kt(j,i) * dexp((e_level(j)-e_level(i))/kt)
     end if  
   end do
 end do
end

subroutine initfiles
 use casc
 implicit none
 character*300 buffer
 integer iostats, i
!
 print*
 call getarg(1,BUFFER)
 if (buffer.eq.'') then	 
   write(*,'(A30)', advance='NO') 'Initfile           : '
   read(*,*) init_file
 else 
   init_file=buffer(1:len_trim(buffer))
 end if 	 
!
 print*, 'Initfile: ',init_file
 open(unit=13,file=init_file,status='unknown')
  read(13,*) rel_path
  write(*,*) 'Relative path      : ', trim(rel_path)
  read(13,*) decay_file
  write(*,*) 'Level scheme file  : ', trim(decay_file)
  decay_file = trim(rel_path) // '/' // trim(decay_file)
  read(13,*) time_cut
  write(*,*) 'Isomer time cut    : ', time_cut, ' s'
  lambda_cut = dlog(2.d0)/time_cut
 close(13)
 open(unit=13,file='element_symbols.txt',status='old')
   iostats=0
   i=0
   element_names(0) = 'NN'
   do while (iostats.eq.0)
     i=i+1
     read(13,*,iostat=iostats) element_names(i)
   end do 
 close(13)   
 
end

subroutine initdecay
 use casc
 implicit none
 integer level
!
 call initlevels
 allocate (lambda_transition(number_levels,number_levels))
 allocate (lambda_transition_effective(number_levels,number_levels))
 allocate (lambda_kt(number_levels,number_levels))
 allocate (lambda_decay(number_levels,number_reaction_types))
 allocate (lambda_decay_effective(number_levels,number_reaction_types))
!
 allocate (output_unit_decay(0:number_levels,0:number_reaction_types))
 allocate (output_unit_transition(number_levels,number_levels))
!
 do level=1,number_levels
   if (lambda(level) .le. lambda_cut) then
     number_isomers      = number_isomers + 1
   else
     number_intermediate = number_intermediate + 1
   end if  
   lambda_transition(level,:) = lambda(level) * transition(level,:)
   lambda_decay(level,:)      = lambda(level) * decay(level,:)
 end do
 print*,'Number of long-lived states  :', number_isomers
 print*,'Number of intermediate states:', number_intermediate
 max_kt = 0.
end 

subroutine initlevels
 use casc
 implicit none
 integer i,j, reaction_id, dz, dn, level_id
 double precision level_energy, level_j, t12, branching, norm, level_p
 integer nbr_target_levels
!
 open(unit=12,file=decay_file,status='old')
   read(12,*) z, n
   a = z+n
   print *,'A, Z, N: ', a, z, n 
   read(12,*) number_reaction_types
   print *,'Number of reaction typs: ', number_reaction_types
   if (number_reaction_types.gt.0) then
     allocate (z_product(number_reaction_types))
     allocate (n_product(number_reaction_types))
     allocate (e_product(number_reaction_types))
     do i=1,number_reaction_types
       read(12,*) reaction_id, dz, dn, level_energy
       z_product(reaction_id) = z + dz
       n_product(reaction_id) = n + dn
       e_product(reaction_id) = level_energy
!       print*,reaction_id, z_product(reaction_id), n_product(reaction_id), level_energy
     end do
     read(12,*) number_levels
	 print*,'Number of levels:', number_levels
	 allocate (e_level(number_levels))
	 allocate (g_level(number_levels))
	 allocate (j_level(number_levels))
	 allocate (p_level(number_levels))
	 allocate (lambda(number_levels))
	 allocate (transition(number_levels,number_levels))
	 allocate (decay(number_levels,number_reaction_types))
     allocate (lambda_transition_wk(number_levels,number_levels))
     do i=1,number_levels
	   read(12,*) level_id, level_energy, level_j, level_p, t12, nbr_target_levels
       e_level(i) = level_energy
       j_level(i) = level_j
       p_level(i) = level_p
       g_level(i) = 2.*j_level(level_id) + 1.	   
       lambda(i)  = dlog(2.d0) / t12
       norm = 0.
       do j=1, nbr_target_levels
         read(12,*) reaction_id, level_id, branching
           if (reaction_id .eq. 0) then  !  IT
             transition(i,level_id) = branching
           else if  (reaction_id .eq. 1) then  !  Reaction changing N,Z
             decay(i,level_id)      = branching
           else
             print*,'something wrong with input. Type can only be 0 or 1, not ',  reaction_id
             stop
           end if
           norm = norm + branching
       end do
       if (norm.eq.0.) then
         print*,'Something wrong with branchings - sum equals zero for level:', i
       else  
         transition(i,:) = transition(i,:) / norm
         decay(i,:)      = decay(i,:) / norm
       end if  
     end do
   end if
 close(12)
 do i=2, number_levels
   if (e_level(i).le.e_level(i-1)) then
     print*,'Please sort levels according to energy.'
     stop
   end if  
 end do 
end

subroutine init_wk
 use casc
 implicit none
 integer i,j
 double precision norm
!
 do i=1,number_levels
   do j=i-1,1,-1
     if (transition(i,j) .eq. 0.d0) then
       call get_weisskopf(i,j)
       print*,'Will assume Weisskopf approximation for transition of state', i, ' to state',j
       print*,'WK - half life time:', dlog(2.d0)/lambda_transition_wk(i,j), ' s'
       lambda_transition(i,j) = lambda_transition_wk(i,j)
       norm = sum(lambda_transition(i,:)) + sum(lambda_decay(i,:))
       transition(i,:) = lambda_transition(i,:)/norm
       print*,'New Branching:', transition(i,j)
     end if
   end do
 end do
end 

subroutine get_weisskopf(i,j)
 use casc
 implicit none
 integer i,j, reaction_id, dz, dn, level_id, ll, hl
 double precision energy,l, h, c, alpha, mp, wk, df, hwz,r, r0
 logical l_even, dp, ttype
!
 ll = min(i,j)
 hl = max(i,j)
 h     = 6.58211d-22                     ! MeV s
 c     = 299792458.d0                    ! m/s
 alpha = 1./137.                         ! cgs
 mp    = 938.3/c**2                      ! MeV*s^2/m^2
 r0    = 1.26d0                          ! fm
 r = r0*(a**(1.d0/3.D0))/1.d15           ! m
 energy = ( e_level(hl) - e_level(ll) )  / 1.d3 
 l      = dabs(j_level(hl) - j_level(ll))
 if (l.gt.0) then
! print*,'Transition energy:', energy, ' MeV'
 l_even = (modulo(l,2.d0) .eq. 0.d0)
 dp     = (p_level(ll)*p_level(hl) .gt. 0.d0)      				! dp = true, if no parity change, dp = false if parity change
 ttype  = (l_even .and. dp ) .or. (.not.l_even .and. .not.dp)	! dp = true if E-transition, dp = false if M-transition
 if (ttype) then
!   print*,'Transition multipolarity:  E', int(l)
   wk = h*((2.*(l+1.))/(l*(df(2.*l+1.))**2))*(9./(l+3.)**2)*((energy/(h*c))**(2.*l+1.))* &
        alpha*c*(R**(2.*l)) ! MeV
 else 
!   print*,'Transition multipolarity:  M', int(l)
   wk = h*((20.d0*(l+1.))/(l*(df(2.*l+1.))**2))*(9./(l+3.)**2)*((energy/(h*c))**(2.*l+1.))* &
        alpha*c*(R**(2.*l-2.))*((h/(mp*c))**2) ! MeV
 end if
! print*,'Weisskopf unit: ', wk,' MeV'
! hwz = dlog(2.d0)*h/wk
! print*,'Half life time: ', hwz, ' s'
 lambda_transition_wk(i,j) = wk/h
 else
   lambda_transition_wk(i,j) = 0.
 end if  
end

recursive double precision function df(n) result(res)
implicit none
double precision n
!
if (n.lt.2) then
  res = 1.d0
else
  res = n*df(n-2.d0)
end if
end      

